#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef JFB_WITH_LIBHANGUL
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <locale.h>
#include <langinfo.h>
#include <sys/ioctl.h>
#include "term.h"
#include "conv.h"
#include "hinput.h"
#include "message.h"

THinput gHin;

int
get_shift_state(void)
{
	int na = 6, ns = 0;
	if(ioctl(fileno(stdin), TIOCLINUX, &na) == 0) ns = na;
	return ns;
}

int
hinput_euckr_filter(ucschar *preedit_ucs4, ucschar cho, ucschar jung, ucschar jong, void *data)
{
	int ret;
	u_char preedit[CONV_SIZE] = {0, };
	/* direct iconv from ucs4 to euc-kr doesn't work. */
	ret = CONV_UCS4_TO_UTF8((char *)preedit_ucs4, preedit);
	if(ret == -1) {
		return 0;
	}
	ret = CONV_UTF8_TO_EUCKR(preedit, preedit);
	if(ret == -1) {
		return 0;
	}
	return 1;
}

void
hinput_init(THinput *h)
{
#ifdef JFB_ENABLE_CURSOR_BLINK
	h->cursor_blink = 0;
#endif
	h->output_encoding = HINPUT_ENCODING_NONE;
	h->preedit_encoding = HINPUT_ENCODING_NONE;

	h->disable_utf8 = 0;
	h->disable_euckr = 0;
	h->hangul_keyboard = HANGUL_KEYBOARD_2;
	h->english_keyboard = ENGLISH_KEYBOARD_QWERTY;
	h->hic = hangul_ic_new(h->hangul_keyboard);
	hangul_ic_process(h->hic, 0);
}

void
hinput_print_status(THinput *h)
{
	print_message("(UXP Hangul) Input encoding:   ");
	if(h->output_encoding == HINPUT_ENCODING_EUCKR) {
		if(h->disable_euckr == 1) {
			print_message("disabled (EUC-KR)");
		}
		else {
			print_message("EUC-KR");
		}
	}
	else if(h->output_encoding == HINPUT_ENCODING_UTF8) {
		if(h->disable_utf8 == 1) {
			print_message("disabled (UTF-8)");
		}
		else {
			print_message("UTF-8");
		}
	}
	else {
		print_message("none");
	}
	print_message("\n");

	print_message("(UXP Hangul) Hangul keyboard:  ");
	if(h->hangul_keyboard == HANGUL_KEYBOARD_2)
		print_message("Dubeolsik");
	else if(h->hangul_keyboard == HANGUL_KEYBOARD_32)
		print_message("Sebeolsik 2");
	else if(h->hangul_keyboard == HANGUL_KEYBOARD_3FINAL)
		print_message("Sebeolsik final");
	else if(h->hangul_keyboard == HANGUL_KEYBOARD_390)
		print_message("Sebeolsik 390");
	else if(h->hangul_keyboard == HANGUL_KEYBOARD_3NOSHIFT)
		print_message("Sebeolsik no-shift");
	else if(h->hangul_keyboard == HANGUL_KEYBOARD_3YETGUL)
		print_message("Sebeolsik yetgul");
	else
		print_message("none");
	print_message("\n");

	print_message("(UXP Hangul) English keyboard:  ");
	if(h->english_keyboard == ENGLISH_KEYBOARD_DVORAK)
		print_message("dvorak");
	else
		print_message("qwerty");
	print_message("\n");
}

int
hinput_read_encoding(char *enc)
{
	if(!enc)
		return HINPUT_ENCODING_NONE;
	while(*enc == ' ') enc++;
	if(strncasecmp(enc, "EUC-KR", 6) == 0 || strncasecmp(enc, "EUCKR", 5) == 0)
		return HINPUT_ENCODING_EUCKR;
	if(strncasecmp(enc, "UTF-8", 5) == 0 || strncasecmp(enc, "UTF8", 4) == 0)
		return HINPUT_ENCODING_UTF8;
	return HINPUT_ENCODING_NONE;
}
	

void
hinput_change_output_encoding(THinput *h, int output_encoding)
{
	if(output_encoding == HINPUT_ENCODING_EUCKR && h->disable_euckr != 1) {
		hangul_ic_set_filter(h->hic, (HangulICFilter)hinput_euckr_filter, NULL);
		h->output_encoding = HINPUT_ENCODING_EUCKR;
	}
	else if(output_encoding == HINPUT_ENCODING_UTF8 && h->disable_utf8 != 1) {
		hangul_ic_set_filter(h->hic, (HangulICFilter)NULL, NULL);
		h->output_encoding = HINPUT_ENCODING_UTF8;
	}
	else {
		hangul_ic_set_filter(h->hic, (HangulICFilter)NULL, NULL);
		h->output_encoding = HINPUT_ENCODING_NONE;
	}
}

void
hinput_change_preedit_encoding(THinput *h, int preedit_encoding)
{
	h->preedit_encoding = preedit_encoding;
}

int
hinput_read_disable(char *disable)
{
	if(!disable)
		return 0;
	else if(strcasecmp(disable, "disable") == 0)
		return 1;
	return 0;
}

void
hinput_disable_utf8(THinput *h, int disable)
{
	if(disable == 1 && h->output_encoding == HINPUT_ENCODING_UTF8) {
		hinput_change_output_encoding(h, HINPUT_ENCODING_NONE);
		hinput_change_preedit_encoding(h, HINPUT_ENCODING_NONE);
	}
	h->disable_utf8 = disable;
}

void
hinput_disable_euckr(THinput *h, int disable)
{
	if(disable == 1 && h->output_encoding == HINPUT_ENCODING_EUCKR) {
		hinput_change_output_encoding(h, HINPUT_ENCODING_NONE);
		hinput_change_preedit_encoding(h, HINPUT_ENCODING_NONE);
	}
	h->disable_euckr = disable;
}

int
hinput_read_hangul_keyboard(char *hkbd)
{
	if(!hkbd)
		return HANGUL_KEYBOARD_2;
	else if(strcasecmp(hkbd, "32") == 0)
		return HANGUL_KEYBOARD_32;
	else if(strcasecmp(hkbd, "3final") == 0 ||
		strcasecmp(hkbd, "3") == 0)
		return HANGUL_KEYBOARD_3FINAL;
	else if(strcasecmp(hkbd, "390") == 0)
		return HANGUL_KEYBOARD_390;
	else if(strcasecmp(hkbd, "3noshift") == 0)
		return HANGUL_KEYBOARD_3NOSHIFT;
	else if(strcasecmp(hkbd, "3yetgul") == 0)
		return HANGUL_KEYBOARD_3YETGUL;
	return HANGUL_KEYBOARD_2;
}

void
hinput_change_hangul_keyboard(THinput *h, int hangul_keyboard)
{
	h->hangul_keyboard = hangul_keyboard;
	hangul_ic_set_keyboard(h->hic, h->hangul_keyboard);
}

int
hinput_read_english_keyboard(char *ekbd)
{
	if(strcasecmp(ekbd, "dvorak") == 0)
		return ENGLISH_KEYBOARD_DVORAK;
	return ENGLISH_KEYBOARD_QWERTY;
}

void
hinput_change_english_keyboard(THinput *h, int english_keyboard)
{
	h->english_keyboard = english_keyboard;
}

void
hinput_exit(THinput *h)
{
	if(h->hic == NULL) {
		return;
	}
	hangul_ic_delete(h->hic);
}

void
hinput_exit_at_exit(void)
{
	hinput_exit(&gHin);
}

void
hinput_try_commit(TTerm *p, THinput *h)
{
	u_char commit[CONV_SIZE] = {0, };
	ucschar *commit_ucs4;
	int ret = 0;

	if(h->hic == NULL) {
		return;
	}

	commit_ucs4 = hangul_ic_get_commit_string(h->hic);
	if(h->output_encoding == HINPUT_ENCODING_UTF8) {
		ret = CONV_UCS4_TO_UTF8((char *)commit_ucs4, commit);
		if(ret == -1) {
			return;
		}
	}
	else if(h->output_encoding == HINPUT_ENCODING_EUCKR) {
		ret = CONV_UCS4_TO_UTF8((char *)commit_ucs4, commit);
		if(ret == -1) {
			return;
		}
		ret = CONV_UTF8_TO_EUCKR(commit, commit);
		if(ret == -1) {
			return;
		}
	}
	write(p->ptyfd, commit, ret);

	return;
}

void
hinput_try_preedit(TTerm *p, THinput *h)
{
	u_char preedit[CONV_SIZE] = {0, };
	ucschar *preedit_ucs4;
	int ret = 0;

	if(h->hic == NULL) {
		return;
	}

	preedit_ucs4 = hangul_ic_get_preedit_string(h->hic);
	if(h->preedit_encoding == HINPUT_ENCODING_UTF8) {
		ucschar *tmp;
		tmp = preedit_ucs4;
		while(*tmp != 0) {
			h->preedit_str[ret] = preedit_ucs4[ret] & 0xFFFFFF;
			ret++;
			tmp++;
			if(ret > 64) { // size defined in libhangul
				break;
			}
		}
		h->preedit_str[ret] = 0;
	}
	else if(h->preedit_encoding == HINPUT_ENCODING_EUCKR) {
		int i;
		/* direct iconv from ucs4 to euc-kr doesn't work. */
		ret = CONV_UCS4_TO_UTF8((char *)preedit_ucs4, preedit);
		if(ret == -1) {
			return;
		}
		ret = CONV_UTF8_TO_EUCKR(preedit, preedit);
		if(ret == -1) {
			return;
		}
		ret = ret / 2;
		for(i = 0; i < ret; i++) {
			h->preedit_str[i] = ((preedit[i * 2] & 0x7F) << 8)
						| (preedit[i * 2 + 1] & 0x7F);
		}
	}
	if(ret > 0) {
		tvterm_refresh(&(p->vterm));
	}
	return;
}

int
hinput_process(TTerm *p, THinput *h, u_char *buf, int ret)
{
	static u_int hstate = 0;
	u_int i, c;

	if(h->hic == NULL) {
		return 0;
	}

	if(buf[0] == KEY_ESC && hstate) {
		hangul_ic_flush(h->hic);
		hinput_try_commit(p, h);
		hstate = 0;
	}
	if(ret == 1 && buf[0] == KEY_SPACE && get_shift_state()) {
		if(hstate) {
			hangul_ic_flush(h->hic);
			hinput_try_commit(p, h);
			hstate = 0;
		}
		else
			hstate = 1;
	}
	else if(hstate) {
		for(i = 0; i < ret; i++) {
			c = buf[i];
			if(h->english_keyboard == ENGLISH_KEYBOARD_DVORAK) {
				c = hangul_ic_dvorak_to_qwerty(c);
			}
			if(c == KEY_BACKSPACE) {
				if(hangul_ic_backspace(h->hic)) {
					tvterm_refresh(&(p->vterm));
				}
				else {
					write(p->ptyfd, &c, 1);
				}
				continue;
			}
			if(hangul_ic_process(h->hic, c) == 0) {
				hinput_try_commit(p, h);
				write(p->ptyfd, &c, 1);
				continue;
			}
			hinput_try_commit(p, h);
		}
		hinput_try_preedit(p, h);
	}
	else
		return 0;
	return 1;
}
#endif /* JFB_WITH_LIBHANGUL */
